﻿param(
###############################################################################
# FOPTracker (FilesystemOperationTracker)
###############################################################################
# To end this script, you can send it CTRL-C or just terminate its process
#
# - comes wih 3 "consumers" (1 for console output, 2 to create redo-scripts)
# - tracked locations can be "offline", tracker will wait/continue/restart
# - probably has bugs and a lot of potential for enhancements
###############################################################################

	#the location to track
		[string] $Location       = "D:",

	#custom id for fopt instance
		[string] $Name           = "Default",

	#filter for the .net filesystemwatcher (don't change unless you know..)
		[string] $Filter         = "*",

    #items (paths or names) to include in json notation like [".*\\.txt", ".*\\logfiles\\.*"]
		[string] $Include        = "",

    #items (paths or names) to exclude in json notation like [".*\\.txt", ".*\\dropbox\\.*"]
		[string] $Exclude        = "['.*dropbox.*']",

	#use the console consumer, echos to console what happens
		[string] $ConOut         = "$true",

	#use the DOpus consumer, echos DOs script console what happens
		[string] $DOpus          = "$false",

	#use the redo-batch-creator consumer. it creates batch files that allow
	#redo operations later & elsewhere (rename, del and move operations)
		[string] $ReDoBatch      = "$false",

	#filename to use for the redo-batch (will reside in the tracked location by default)
		[string] $ReDoBatchFilePath  = ".fop_redobatch.bat",

	#use the redo-batch-creator consumer. it creates batch files that allow
	#redo operations later & elsewhere (rename, del and move operations)
		[string] $ReDoPoSh       = "$true",

	#filename to use for the redo-script (will reside in the tracked location by default)
		[string] $ReDoPoShFilePath  = ".fop_redoposh.ps1",

	#minimum filesize in bytes to create redo-entries for (1048576 = 1MB)
		[int32]  $ReDoMinSize    = 1048576,

	#track and create redo-script for changes on folders
		[string] $ReDoFolders    = "$true",

	#track and create redo-script for changes on files
		[string] $ReDoFiles      = "$true",

	#enable debug output
		[string] $Debug          = "$false",

	#param sets in json notation like this:
    #"[ {ID:'System',Location:'C:\\',Filter:'*.txt'}, {ID:'Data',Location:'D:\\'} ]"
	#make sure to use double backslashes in paths! (think javascript here)
		$sets          = $null
)
###############################################################################
if (!$sets -and $env:computername -like "daily*"){ #
	$sets = "[{ Name:'System',Location:'C:\\ProgramData\\GPSoftware\\Directory Opus',Filter:'*.*'}, {Name:'Data',Location:'D:\\'}, {Location:'T:\\'} ]";
	$sets = "[{ Name:'D:\\dat', Location:'D:\\dat', DOpus:true, Include:['.*\\.txt'], Exclude:['Drop\\\\.*', 'Dropbox\\\\.*'] }]";
    $sets = "[{ Name:'D:\\dat', Location:'D:\\dat', DOpus:true }]";
}
###############################################################################
$version                = "0.2";
$date                   = "2o16-o2-o2";
###############################################################################
function ObjectToHashtable($o){
###############################################################################
	$ht = @{};
	if ($o.gettype().Fullname -like "*PSCustomObject"){
		$o.psobject.properties | foreach { $ht[$_.Name] = $_.Value }}
	if ($o.gettype().Fullname -eq "System.Object[]"){
		$o.psobject.properties | foreach { $ht[$_.Name] = $_.Value }}
	if ($o.gettype().Fullname -like "*Hashtable"){
		return $o;}
	return $ht;
}
###############################################################################
function ToBoolean($v){ #v0.2
###############################################################################
	if (($v -eq $false) -or ($v -like '*false') -or ($v -eq 0) -or ($v -eq '0') -or ($v -like 'no') ){
		return ([boolean] $false);}
	if (($v -eq $true)  -or ($v -like '*true') -or  ($v -eq 1) -or ($v -eq '1') -or ($v -eq 'yes') ){
		return ([boolean] $true);}
	throw "Cannot convert [$v] to boolean."
}
###############################################################################
[bool]  $ConOut             = ToBoolean $ConOut;
[bool]  $Debug              = ToBoolean $Debug;
[bool]  $ReDoBatch          = ToBoolean $ReDoBatch;
[bool]  $ReDoPoSh           = ToBoolean $ReDoPoSh;
[bool]  $ReDoFolders        = ToBoolean $ReDoFolders;
[bool]  $ReDoFiles          = ToBoolean $ReDoFiles;
[bool]  $DOpus              = ToBoolean $DOpus;
[array] $Include            = @( (ConvertFrom-Json $Include) );
[array] $Exclude            = @( (ConvertFrom-Json $Exclude) );
###############################################################################
$ErrorActionPreference     = 'stop';
$void                      = set-strictmode -version latest;
$void                      = $error.clear();
$void                      = set-location $PSScriptRoot;
$IsISE                     = (Test-Path Variable:\psISE);
if (!$IsISE)                 {$Host.UI.RawUI.WindowTitle=$myinvocation.mycommand.name;}
###############################################################################
function PrepareParamSets($sets){
###############################################################################
    if ($sets -eq $null){
		$sets = @( @{Location = $global:Location} ); } #single set mode
	else {
		$sets = @( (ConvertFrom-Json $sets) ); } #multi set mode

	$params = "Name","Filter","ReDoBatch","RedoBatchFilePath","RedoPoSh","ReDoPoShFilePath","DOpus","ReDoFolders","ReDoMinSize","ReDoFolders","ReDoFiles","ConOut","Debug";
	for($i=0;$i -lt $sets.length;$i++){
		$set = [hashtable] (ObjectToHashtable $sets[$i]);
        
		foreach($param in $params){
			if (!$set.ContainsKey($param)){
                $set[$param] = Get-Variable -ValueOnly -Scope global $param; }
		}
        if (!$set.ContainsKey("Name")){ $set["Name"] = "Tracker"+$i;}
		if (!$set.ContainsKey("Location")){
			write-host "E   Bad param set #$($i+1), name [$($set.name)] - Location key is missing."; exit; }
		#write-host "ParamSet #$((''+($i+1)).PadLeft(2,"0")): name [$($set.name.PadRight(10))], location [$($set.location)]";
		$sets[$i] = $set;
	}
	return $sets;
}
###############################################################################
$block = { param($scriptPath, $set, $setNumber)
###############################################################################
	function LogOut($text){
		Write-Verbose -Message $text -Verbose;
	}
	."$scriptPath\include\NewFOPTracker.ps1";
	."$scriptPath\include\NewFOPTConsumer_ConOut.ps1";
    ."$scriptPath\include\NewFOPTConsumer_DOpus.ps1";
	."$scriptPath\include\NewFOPTConsumer_ReDoBatchCreator.ps1";
	."$scriptPath\include\NewFOPTConsumer_ReDoPowershell.ps1";
	$set.fopt = NewFOPTracker $set.Name $set.Location $set.Filter $set.Include $set.Exclude;
    $set.fopt.Debug = $set.Debug;
	"Preparing FOPTracker v$($set.fopt.__Version) for set #$setNumber..";
	$set.fopt.Init();
	if ($set.ConOut)    { $set.fopt.AddConsumer( (NewFOPTConsumer_ConOut) ); }
    if ($set.DOpus)     { $set.fopt.AddConsumer( (NewFOPTConsumer_DOpus) ); }
	if ($set.RedoBatch) { $set.fopt.AddConsumer( (NewFOPTConsumer_ReDoBatchCreator $set.ReDoBatchFilePath $set.RedoMinSize $set.ReDoFolders $set.ReDoFiles) ); }
	if ($set.RedoPoSh)  { $set.fopt.AddConsumer( (NewFOPTConsumer_ReDoPowershell   $set.ReDoPoShFilePath $set.RedoMinSize $set.ReDoFolders $set.ReDoFiles) ); }
	$set.fopt.Initialized;
	$set.fopt.Start();
}
###############################################################################
function LaunchTrackers($sets, $jobs){
###############################################################################
    for($i=0;$i -lt $sets.count;$i++){
		$set = $sets[$i];
		$setNumber = $($i+1);
        write-host "Launching tracker #$((''+($i+1)).PadLeft(2,"0")): name [$($set.name.PadRight(10))], location [$($set.location)]..";
		#&$ScriptBlock $psscriptroot $set $setNumber;
		#&$ScriptBlock.Invoke(@($psscriptroot,$set,$setNumber));

		$job = [powershell]::Create().AddScript($block).AddArgument($psscriptroot).AddArgument($set).AddArgument($setNumber);
		$job.RunspacePool = $runspacePool;
		$jobs += New-Object PSObject -Property @{
			Name = $set.Name; #("SetJob"+$i);
			Pipe = $job;
			Result = $job.BeginInvoke();
		}
	}
    return $jobs;
}
###############################################################################
function WaitAndEchoPipeData($jobs){
###############################################################################
    do {
		start-sleep -seconds 1;
		foreach ($job in $jobs){
			#write-host $Job.Result;
			if ($job.Pipe.Streams.Error.Count){
				write-host "Error ($($job.Name)):";
				foreach($m in $Job.Pipe.Streams.Error.GetEnumerator()){
					write-host $m;
				}
				$job.Pipe.Streams.Error.Clear();
			}
			#write-host $Job.Pipe.Streams.Progress
			if ($job.Pipe.Streams.Verbose.Count){
				foreach($m in $Job.Pipe.Streams.Verbose.GetEnumerator()){
					write-host $m;
				}
				$job.Pipe.Streams.Verbose.Clear();
			}
			#write-host $Job.Pipe.Streams.Warning
		}
	} while ( $jobs.Result.IsCompleted -contains $false )
}
###############################################################################
function StopTrackers($jobs){
###############################################################################
    write-host
	write-host "Stopping all trackers..";
	$results = @()
	foreach ($job in $jobs ){
		write-host "Stopping/Removing [$($job.Name)]";
		$job.Pipe.Stop();
		$job.Pipe.Dispose();
		#$result = $Job.Pipe.EndInvoke($Job.Result)
		#write-host $result
		#$Job.Dispose();
	}
}
###############################################################################
# main
###############################################################################
write-host "FOPTracker v$version $date, tb (email: mb_foptracker@steax.net)";
try {
    $sets = @(PrepareParamSets $sets);
    $jobs = @()
    $maxThreads = $sets.count;
    $runspacePool = [RunspaceFactory ]::CreateRunspacePool(1, $maxThreads)
    $runspacePool.Open();

    $jobs = LaunchTrackers $sets $jobs;
	write-host "Waiting for CTRL-C to exit.."
	write-host
	WaitAndEchoPipeData $jobs;
	write-host "All trackers completed."
}
catch{
    write-host "An error occured, see stacktrace for more..";
    throw $error;
}
finally{
    #will be triggered by ctrl-c
    StopTrackers $jobs;
	$runspacePool.Close();
	#$results;
}

###############################################################################
#global trap for severe exceptions, exit()s and error situations
###############################################################################
trap {
	set-location $PSScriptRoot;
	write-host;
	###########################################################################
	function FormatErrorStack( $error) {
		$eStackOut = "";
		for($e=$error.count-1;$e -gt -1 ; $e--){
				$eIndex = ($error.count-1-$e);
				$eStackOut += "----------------------------------------------------------------------------< "+([string] $eIndex).PadLeft(2,'0')+"`n";
				$eStackOut += "Msg  : $($error[$e].ToString())"+"`n";
				$eStackOut += "File : $($error[$e].InvocationInfo.ScriptName)"+"`n";
				$eStackOut += "Pos  : $($error[$e].InvocationInfo.ScriptLineNumber) $($error[$e].InvocationInfo.OffsetInLine)"+"`n";
				$eStackOut += "Line : $($error[$e].InvocationInfo.Line.trim())"+"`n";
				if (($e-1) -gt -1){ $eStackOut += "`n"; }
		}
		return $eStackOut;
	}
	###########################################################################
	write-host "Error trap:"
	$errorText = FormatErrorStack $error;
	write-host $errorText;

	exit 10;
}
