﻿###############################################################################
function NewFOPTConsumer_ReDoBatchCreator(  $fileName  = ".fop_redobatch.bat",
											$minSize   = 1024000,
											$doFolders = $true,
											$doFiles   = $true){
###############################################################################
	$fs=@(); #for storing the function/method script blocks
	$me= #for storing the objects properties
	###########################################################################
	@{
		__Version        = 0.1;

		FOPTracker       = $null;
		Prefix           = ""; #[TrackerName]

		FileName         = "";
		MinSize          = 1024000;
		DoFolders        = $true;
		DoFiles          = $true;

		BatchFilePath    = "";
	}
	###########################################################################
	function constructor($fileName, $minSize, $doFolders, $doFiles){
	###########################################################################
		$this.FileName  = $fileName;
		$this.DoFolders = $doFolders;
		$this.DoFiles   = $doFiles;
		$this.MinSize   = $minSize;
	}
	###########################################################################
	$fs+=,@('_Init', { param( $FOPTracker )
	###########################################################################
		$this.FOPTracker   = $FOPTracker;
		#not using join path, it requires paths to already exist (wtf!?)
		$this.BatchFilePath = $this.FOPTracker.Folder.TrimEnd("\/")+"\"+$this.FileName;
		if ($this.FileName -like "?:*" -or $this.FileName -like "\\*"){
			$this.BatchFilePath = $this.FileName;}
		$this.FileName      = split-path -leaf $this.FileName;
		$this.Prefix      = $this.FOPTracker.LogPrefix+".ReDoBat: ";
		LogOut "    ReDo: Creating/appending to [$($this.BatchFilePath)]";
	});
	###########################################################################
	$fs+=,@('AddBatchCommand', { param( $cmd, $op )
	###########################################################################
		#create head-line for new redo-batch, telling the location/folder
		if (![System.IO.File]::Exists($this.BatchFilePath)){
			[System.IO.File]::WriteAllText($this.BatchFilePath,
				"REM foptracker redo, location: `"$($this.FOPTracker.Folder)`""+[Environment]::NewLine,
				[System.Text.Encoding]::UTF8);
		}

		$cmd = $cmd -replace "%", "%%";
		#append command
		[System.IO.File]::AppendAllText($this.BatchFilePath,
			$cmd + [Environment]::NewLine,
			[System.Text.Encoding]::UTF8);
	});
	###########################################################################
	$fs+=,@('FilterItem', { param( $path, $opType )
		if (!$this.FOPTracker.Initialized){ return; }
		if ($path    -like "*\$($this.FileName)*" -or
			$path    -like "$($this.FileName)*" -or
			$path    -like '*\$RECYCLE.BIN*'){
			LogOut "$($this.Prefix)Skipped [$path], special location/file ($opType)";
			return $false;
		}

		if ($opType -eq "deleted"){
			#stop filtering, since we have not enough details (just the path)
			#and also can't get information for deleted items
			return $true;
		}

		$item = get-item -LiteralPath $path -ErrorAction Ignore;
		if (!$item){ return $false;}

		$isFile = $false;
		if (($item.gettype()).Name -eq "FileInfo"){ $isFile = $true; }

		if ($isFile -and !$this.DoFiles) {
			LogOut "$($this.Prefix)Skipped [$path], not a directory ($opType)";
			return $false;
		}
		if (!$isFile -and !$this.DoFolders) {
			LogOut "$($this.Prefix)Skipped [$path], not a file [$path] ($opType)";
			return $false;
		}
		if ($isFile -and $item.length -lt $this.MinSize) {
			LogOut "$($this.Prefix)Skipped [$path], too small ($opType)";
			return $false;
		}
		return $true;
	});
	###########################################################################
	$fs+=,@('OnCreate', { param( $op )
		#LogOut "$($this.Prefix)CREATED:" $op.filePath;
	});
	###########################################################################
	$fs+=,@('OnChange', { param( $op )
		#LogOut "$($this.Prefix)CHANGED:" $op.filePath;
	});
	###########################################################################
	$fs+=,@('OnRename', { param( $op )
		if (!$this.FilterItem( $op.fullPath, $op.opType )){ return; }
		$cmd = "rename `"$($op.filePathOld)`" `"$($op.name)`"";
		$this.AddBatchCommand("echo "+$this.FOPTracker.ID+" "+$cmd, $op);
		$this.AddBatchCommand($cmd, $op);
		LogOut "$($this.Prefix)Renamed [$($op.filePathOld)] -> [$($op.filePath)]";
	});
	###########################################################################
	$fs+=,@('OnMove', { param( $opDelete, $opCreate, $opMerged )
		if (!$this.FilterItem( $opCreate.fullPath, "moved" )){ return; }
		$cmd = "move `"$($opDelete.filePath)`" `"$($opCreate.filePath)`"";
		$this.AddBatchCommand("echo "+$cmd, $op);
		$this.AddBatchCommand($cmd, $op);
		LogOut "$($this.Prefix)Moved   [$($opDelete.filePath)] -> [$($opCreate.filePath)]";
	});
    ###########################################################################
	$fs+=,@('OnReplace', { param( $op)
        #LogOut "$($this.Prefix)Replacd [$($op.filePath)]";
	});
	###########################################################################
	$fs+=,@('OnDelete', { param( $op )
		if (!$this.FilterItem( $op.fullPath, $op.opType )){ return; }
		$cmd = "del /F /S /Q `"$($op.filePath)`"";
		$this.AddBatchCommand("echo "+$cmd, $op);
		$this.AddBatchCommand($cmd, $op);
		LogOut "$($this.Prefix)Deleted [$($op.filePath)]";
	});
	###########################################################################
	$me.__Type = (gv MyInvocation -Scope 0).Value.MyCommand.Name -replace "^NEW-?", "";
	if (!$me.ContainsKey("__Version")){ $me.__Version = 0;}
	$this=new-object -typename 'System.Management.Automation.PSObject' -property $me;
	foreach($function in $fs){
		$this | add-member -me scriptmethod -na $function[0] -va $function[1]; }
	###########################################################################
	constructor $local:fileName $local:minSize $local:doFolders $local:doFiles;
	return $this;
}