﻿###############################################################################
function NewFOPTConsumer_ReDoPowershell(  $fileName  = ".fop_redoposh.ps1",
											$minSize   = 1024000,
											$doFolders = $true,
											$doFiles   = $true){
###############################################################################
	$fs=@(); #for storing the function/method script blocks
	$me= #for storing the objects properties
	###########################################################################
	@{
		__Version        = 0.1;

		FOPTracker       = $null;
		Prefix           = ""; #[TrackerName]

		FileName         = "";
		MinSize          = 1024000;
		DoFolders        = $true;
		DoFiles          = $true;

		PSFilePath       = "";
	}
	###########################################################################
	function constructor($fileName, $minSize, $doFolders, $doFiles){
	###########################################################################
		$this.FileName  = $fileName;
		$this.DoFolders = $doFolders;
		$this.DoFiles   = $doFiles;
		$this.MinSize   = $minSize;
	}
	###########################################################################
	$fs+=,@('_Init', { param( $FOPTracker )
	###########################################################################
		$this.FOPTracker = $FOPTracker;
		#not using join path, it requires paths to already exist (wtf!?)
		$this.PSFilePath  = $this.FOPTracker.Folder.TrimEnd("\/")+"\"+$this.FileName;
		if ($this.FileName -like "?:*" -or $this.FileName -like "\\*"){
			$this.PSFilePath = $this.FileName;}
		$this.FileName      = split-path -leaf $this.FileName;
		$this.Prefix      = $this.FOPTracker.LogPrefix+".ReDoPS : ";
		LogOut "$($this.Prefix)Creating/appending to [$($this.PSFilePath)]";
	});
	###########################################################################
	$fs+=,@('AddCommand', { param( $cmd, $op )
	###########################################################################
		#create head-line for new redo-batch, telling the location/folder
		if (![System.IO.File]::Exists($this.PSFilePath)){

			$head = "#foptracker redo, location: "+ $($this.FOPTracker.Folder) + "`r`n";
			$head +=
@'
###############################################################################
$erroractionpreference="stop";
function spl($p, $m='parent|leaf'){ $parts = @(([string] $p).split("\")); if ($m -like 'p*'){ if ($parts.length -eq 1){ return ""; } return ($parts[0..($parts.length-2)] -join "\"); } return $parts[-1];}
function rf($o,$n){ write-host "Renaming in '$(spl $o)'`r`n" "   from $(spl $o 'l')`r`n" "   to   $n";try{ if ($psversiontable.psversion -like "3*"){ren -for -l $o -n $n;} else {ren -for -pat $o -n $n;}}catch{ write-host "E   FAILED:" $error[0];}}
function mf($f,$d){ write-host "Moving '$(spl $f 'l')'`r`n    from $((spl $f))`r`n    to   $(spl $d)";try{move -for -l $f -dest $(spl $d);}catch{ write-host "E   FAILED:" $error[0];}}
function df($f   ){ write-host "Deleting '$f'";try{del -for -rec -l $f;}catch{ write-host "E   FAILED:" $error[0];}}
###############################################################################
'@
			[System.IO.File]::WriteAllText($this.PSFilePath,
				$head+[Environment]::NewLine,
				[System.Text.Encoding]::UTF8);
		}

		#append command
		[System.IO.File]::AppendAllText($this.PSFilePath,
			$cmd + " #"+(get-date -format "yyyy_MM_dd HH:mm")+[Environment]::NewLine,
			[System.Text.Encoding]::UTF8);
	});
	###########################################################################
	$fs+=,@('FilterItem', { param( $path, $opType )
		if (!$this.FOPTracker.Initialized){ return; }
		if ($path    -like "*\$($this.FileName)*" -or
			$path    -like "$($this.FileName)*" -or
			$path    -like '*\$RECYCLE.BIN*'){
			LogOut "$($this.Prefix)Skipped [$path], special location/file ($opType)";
			return $false;
		}

		if ($opType -eq "deleted"){
			#stop filtering, since we have not enough details (just the path)
			#and also can't get information for deleted items
			return $true;
		}

		$item = get-item -LiteralPath $path -ErrorAction Ignore;
		if (!$item){ return $false;}

		$isFile = $false;
		if (($item.gettype()).Name -eq "FileInfo"){ $isFile = $true; }

		if ($isFile -and !$this.DoFiles) {
			LogOut "$($this.Prefix)Skipped [$path], not a directory ($opType)";
			return $false;
		}
		if (!$isFile -and !$this.DoFolders) {
			LogOut "$($this.Prefix)Skipped [$path], not a file [$path] ($opType)";
			return $false;
		}
		if ($isFile -and $item.length -lt $this.MinSize) {
			LogOut "$($this.Prefix)Skipped [$path], too small ($opType)";
			return $false;
		}
		return $true;
	});
	###########################################################################
	$fs+=,@('OnCreate', { param( $op )
		#LogOut "$($this.Prefix)CREATED:" $op.filePath;
	});
	###########################################################################
	$fs+=,@('OnChange', { param( $op )
		#LogOut "$($this.Prefix)CHANGED:" $op.filePath;
	});
	###########################################################################
	$fs+=,@('OnRename', { param( $op )
		if (!$this.FilterItem( $op.fullPath, $op.opType )){ return; }
		$cmd = "rf `"$($op.filePathOld)`" `"$($op.name)`"";
		$this.AddCommand($cmd, $op);
        LogOut "$($this.Prefix)Renamed [$($op.filePathOld)] -> [$($op.filePath)]";
	});
	###########################################################################
	$fs+=,@('OnMove', { param( $opDelete, $opCreate, $opMerged )
		if (!$this.FilterItem( $opCreate.fullPath, "moved" )){ return; }
		$cmd = "mf `"$($opDelete.filePath)`" `"$($opCreate.filePath)`""
		$this.AddCommand($cmd, $op);
		LogOut "$($this.Prefix)Moved   [$($opDelete.filePath)] -> [$($opCreate.filePath)]";
	});
    ###########################################################################
	$fs+=,@('OnReplace', { param( $op)
        #LogOut "$($this.Prefix)Replacd [$($op.filePath)]";
	});
	###########################################################################
	$fs+=,@('OnDelete', { param( $op )
		if (!$this.FilterItem( $op.fullPath, $op.opType )){ return; }
		$cmd = "df `"$($op.filePath)`"";
		$this.AddCommand($cmd, $op);
		LogOut "$($this.Prefix)Deleted [$($op.filePath)]";
	});
	###########################################################################
	$me.__Type = (gv MyInvocation -Scope 0).Value.MyCommand.Name -replace "^NEW-?", "";
	if (!$me.ContainsKey("__Version")){ $me.__Version = 0;}
	$this=new-object -typename 'System.Management.Automation.PSObject' -property $me;
	foreach($function in $fs){
		$this | add-member -me scriptmethod -na $function[0] -va $function[1]; }
	###########################################################################
	constructor $local:fileName $local:minSize $local:doFolders $local:doFiles;
	return $this;
}