﻿###############################################################################
function NewFOPTracker($name, $folder, $filter="*.*", $include=@(), $exclude=@()){
###############################################################################
	$fs=@(); #for storing the function/method script blocks
	$me= #for storing the objects properties
	###########################################################################
	@{
		__Version        = 0.2;

		Watcher          = $null;
		ID               = 0; #635896967768825812
		Name             = "";
		Folder           = "";
		Filter           = "";
		Include          = @();
		Exclude          = @();

		Debug            = $true;

		Recursive        = $true;
		RaiseEvents      = $true;
		NotifyFilters    =	[System.IO.NotifyFilters]::DirectoryName -bor
							[System.IO.NotifyFilters]::FileName -bor
							[System.IO.NotifyFilters]::LastWrite;

		Initialized      = $false;
		Listening        = $false;

		Ops              = @{};
		OpsToAdd         = @{};
		OpKeysToRemove   = @();
		LogPrefix        = "";

		Consumers        = @();
	}
	###########################################################################
	function constructor($name, $folder, $filter, $include, $exclude){
	###########################################################################
		$this.Name      = $name;
		$this.Folder    = $folder.TrimEnd("\/");
		$this.Filter    = $filter;
		$this.Include   = $include;
		$this.Exclude   = $exclude;
		$this.ID        = (get-date).Ticks;
		$this.LogPrefix = "["+(&{if($name.length -gt 8){$name.substring(0,8);} else{$name.PadRight(8)} })+"]";
	}
	###########################################################################
	$fs+=,@('Init', { param()
	###########################################################################
		LogOut "$($this.LogPrefix) Initializing [$($this.Name) ($($this.ID))]..";
		for($i=0;$i -lt $this.Include.count;$i++){
			$this.Include[$i] = new-object System.Text.RegularExpressions.Regex ($this.Include[$i], [System.Text.RegularExpressions.RegexOptions]::IgnoreCase);
			LogOut "$($this.LogPrefix)     Including [$($this.Include[$i])]";
		}
		for($i=0;$i -lt $this.Exclude.count;$i++){
			$this.Exclude[$i] = new-object System.Text.RegularExpressions.Regex ($this.Exclude[$i], [System.Text.RegularExpressions.RegexOptions]::IgnoreCase);
			LogOut "$($this.LogPrefix)     Excluding [$($this.Exclude[$i])]";
		}
		try{
			$this.Watcher = New-Object IO.FileSystemWatcher $this.Folder, $this.Filter
			$this.Watcher.IncludeSubdirectories = $this.Recursive;
			$this.Watcher.EnableRaisingEvents   = $this.RaiseEvents;
			$this.Watcher.NotifyFilter          = $this.NotifyFilters;
			$this.Initialized = $true;
		} catch {}

	});
	###########################################################################
	$fs+=,@('Start', { param()
	###########################################################################
		LogOut "$($this.LogPrefix)     Starting..";
		if ($this.Initialized){
			LogOut "$($this.LogPrefix)     Un/Registering events..";
			$this.UnRegisterEvents();
			$this.RegisterEvents();
		} else {
			LogOut "$($this.LogPrefix) E   Not initialized";
		}
		$this.Listen();
	});
	###########################################################################
	$fs+=,@('Stop', { param()
	###########################################################################
		LogOut "$($this.LogPrefix)     Stopping..";
		$this.Listening = $false;
		$this.UnRegisterEvents( ($info = $false) );
		$this.Initialized = $false;
		LogOut "$($this.LogPrefix)     Done.";
	});
	###########################################################################
	$fs+=,@('AddConsumer', { param( $consumer)
	###########################################################################
		LogOut "$($this.LogPrefix)     Adding Consumer [$($consumer.__Type)]..";
		$consumer._Init($this);
		$this.Consumers += $consumer;
	});
	###########################################################################
	$fs+=,@('RegisterEvents', { param()
	###########################################################################
		LogOut "$($this.LogPrefix)     Registering events..";
		# CREATE
		#######################################################################
		Register-ObjectEvent $this.Watcher Created -SourceIdentifier ("FileCreated"+$this.ID) -MessageData @{tracker = $this;} -Action {
			$tracker = $event.MessageData.tracker;
			if ($tracker.Debug) { LogOut "C $($event.SourceEventArgs.Name)"; }
			$op = CreateOp "created" $event $tracker;
			if (!$op.included) { if ($tracker.Debug) { LogOut "    Not included"; } return; }
			if ($op.excluded) { if ($tracker.Debug) { LogOut "    Excluded"; } return; }
			$tracker.OpsToAdd.Add( "created:"+$op.name, $op);
		}
		# CHANGE
		#######################################################################
		Register-ObjectEvent $this.Watcher Changed -SourceIdentifier ("FileChanged"+$this.ID) -MessageData @{tracker = $this;} -Action {
			$tracker = $event.MessageData.tracker;
			if ($tracker.Debug) { LogOut "A $($event.SourceEventArgs.Name)"; }
			$op = CreateOp "changed" $event $tracker;
			if (!$op.included) { if ($tracker.Debug) { LogOut "    Not included"; } return; }
			if ($op.excluded) { if ($tracker.Debug) { LogOut "    Excluded"; } return; }
			$tracker.OpsToAdd.Add( "changed:"+$op.name, $op);
		}
		# RENAME
		#######################################################################
		Register-ObjectEvent $this.Watcher Renamed -SourceIdentifier ("FileRenamed"+$this.ID) -MessageData @{tracker = $this;} -Action {
			$tracker = $event.MessageData.tracker;
			if ($tracker.Debug) { LogOut "R $($event.SourceEventArgs.Name) $($tracker.OpsToAdd.Count)"; }
			#$timeStamp = $event.TimeGenerated
			$op = CreateOp "renamed" $event $tracker;
			if (!$op.included) { if ($tracker.Debug) { LogOut "    Not included"; } return; }
			if ($op.excluded) { if ($tracker.Debug) { LogOut "    Excluded"; } return; }
			$tracker.OpsToAdd.Add( "renamed:"+$op.name, $op);
		}
		# DELETE
		#######################################################################
		Register-ObjectEvent $this.Watcher Deleted -SourceIdentifier ("FileDeleted"+$this.ID) -MessageData @{tracker = $this;} -Action {
			$tracker = $event.MessageData.tracker;
			if ($tracker.Debug) { LogOut "D $($event.SourceEventArgs.Name)"; }
			$op = CreateOp "deleted" $event $tracker;
			if (!$op.included) { if ($tracker.Debug) { LogOut "    Not included"; } return; }
			if ($op.excluded) { if ($tracker.Debug) { LogOut "    Excluded"; } return; }
			$tracker.OpsToAdd.Add( "deleted:"+$op.name, $op);
		}
	});
	###########################################################################
	$fs+=,@('UnRegisterEvents', { param($info=$true)
	###########################################################################
		try{ Unregister-Event ("FileCreated"+$this.ID) -ErrorAction stop; } catch{ if (!$info) { LogOut ("I   FileCreated"+$this.ID+" not registered (yet)."); } }
		try{ Unregister-Event ("FileChanged"+$this.ID) -ErrorAction stop; } catch{ if (!$info) { LogOut ("I   FileChanged"+$this.ID+" not registered (yet)."); } }
		try{ Unregister-Event ("FileRenamed"+$this.ID) -ErrorAction stop; } catch{ if (!$info) { LogOut ("I   FileRenamed"+$this.ID+" not registered (yet)."); } }
		try{ Unregister-Event ("FileDeleted"+$this.ID) -ErrorAction stop; } catch{ if (!$info) { LogOut ("I   FileDeleted"+$this.ID+" not registered (yet)."); } }
	});
	###########################################################################
	$fs+=,@('Listen', { param()
	###########################################################################
		LogOut "$($this.LogPrefix)     Listening on [$($this.Folder)]..";
		$this.Listening = $true;
		$waiting = $false;
		try{
			while ($this.Listening){
				#LogOut "Tick"
				while (!$this.Initialized){
					if (!$waiting){
						LogOut "$($this.LogPrefix) Tracker not initialized..";}
					if (![System.IO.Directory]::Exists($this.Folder)){
						if (!$waiting){
							LogOut "$($this.LogPrefix) Folder [$($this.Folder)] missing, waiting..";}
						$waiting = $true;
						sleep 1;
					} else {
						LogOut "$($this.LogPrefix) Folder [$($this.Folder)] appeared, re-initializing.."
						$this.Init();
						$this.Start();
						return;
					}
				}

				if (![System.IO.Directory]::Exists($this.Folder)){
					LogOut "$($this.LogPrefix) Folder [$($this.Folder)] disappeared..";
					$this.Stop();
					$this.Listen();
					return;
				}

				#LogOut "Tick.." $this.Ops.Keys.Count $this.OpKeysToRemove.count $this.OpsToAdd.Keys.Count;

				$this.OpKeysToRemove = @();
				foreach($opKey in $this.Ops.Keys){
					$op = $this.Ops[$opKey];
					if ($op.opProcessed){ continue;}

					#LogOut $op.name $op.type $op.opAge
					$op.opAge++;
					if ($op.opAge -gt 1){
						if ($op.opType -eq "created"){ $this.OnCreate( $op ); }
						if ($op.opType -eq "changed"){ $this.OnChange( $op ); }
						if ($op.opType -eq "renamed"){ $this.OnRename( $op ); }
						if ($op.opType -eq "deleted"){ $this.OnDelete( $op ); }
						$this.Ops[$opKey].opProcessed = $true;
						$this.OpKeysToRemove += $opKey;
						continue;
					}

					$otherOp = "deleted";
					if ($op.opType -eq "deleted") { $otherOp = "created";}
					$otherOpKey = $otherOp+":"+$op.name;

					if ($this.Ops.ContainsKey($otherOpKey)){
						if ($this.Ops[$opKey].opType -eq "created"){
							$opDelete = $this.Ops[$otherOpKey];
							$opCreate = $this.Ops[$opKey];
						} else {
							$opDelete = $this.Ops[$opKey];
							$opCreate = $this.Ops[$otherOpKey];
						}
						if ($opDelete.fullPath.toLower() -eq $opCreate.fullPath.toLower()){
							$opCreate.type = "replaced";
							$this.OnReplace( $opCreate )
						} else {
							$this.OnMove( $opDelete, $opCreate)
						}
						$this.Ops[$opKey].opProcessed = $true;
						$this.Ops[$otherOpKey].opProcessed = $true;
						$this.OpKeysToRemove += $opKey;
						$this.OpKeysToRemove += $otherOpKey;
					}
				}

				#remove processed ops by not looping over the $ops hashmap
				for($i=0;$i -lt $this.OpKeysToRemove.Length;$i++){
					#LogOut "Removing:" $this.OpKeysToRemove[$i];
					$this.Ops.Remove($this.OpKeysToRemove[$i]);
				}

				#add ops from events to main hashmap $ops
				#copy keys, since we cannot alter a hashmap while enumerating
				$keys = [string[]]($this.OpsToAdd.Keys)
				foreach($opKey in $keys){
					$this.Ops[$opKey] = $this.OpsToAdd[$opKey];
					$this.OpsToAdd.Remove($opKey);
				}

				sleep 1
			}
		}
		finally
		{
			if ($this.Initialized){ $this.Stop(); }
		}
	});
	###########################################################################
	$fs+=,@('OnCreate', { param( $op )
		foreach($consumer in $this.Consumers){ $consumer.OnCreate($op); }
	});
	###########################################################################
	$fs+=,@('OnChange', { param( $op )
		foreach($consumer in $this.Consumers){ $consumer.OnChange($op); }
	});
	###########################################################################
	$fs+=,@('OnRename', { param( $op )
		foreach($consumer in $this.Consumers){ $consumer.OnRename($op); }
	});
	###########################################################################
	$fs+=,@('OnMove', { param( $opDelete, $opCreate )
		$opMerged = $opCreate.Clone();
		$opMerged.pathOld = $opDelete.path;
		$opMerged.filePathOld = $opDelete.filePath;
		$opMerged.fullPathOld = $opDelete.fullPath;
		foreach($consumer in $this.Consumers){ $consumer.OnMove($opDelete, $opCreate, $opMerged); }
	});
	###########################################################################
	$fs+=,@('OnReplace', { param( $o)
		foreach($consumer in $this.Consumers){ $consumer.OnReplace($op); }
	});
	###########################################################################
	$fs+=,@('OnDelete', { param( $op )
		foreach($consumer in $this.Consumers){ $consumer.OnDelete($op); }
	});
	###########################################################################
	$me.__Type = (gv MyInvocation -Scope 0).Value.MyCommand.Name -replace "^NEW-?", "";
	if (!$me.ContainsKey("__Version")){ $me.__Version = 0;}
	$this=new-object -typename 'System.Management.Automation.PSObject' -property $me;
	foreach($function in $fs){
		$this | add-member -me scriptmethod -na $function[0] -va $function[1]; }
	###########################################################################
	constructor $local:name $local:folder $local:filter $local:include $local:exclude;
	return $this;
}
###############################################################################
function CreateOP( $type="renamed", $event, $tracker){
###############################################################################
	$op = @{
		opType      = $type;
		opProcessed = $false;
		opAge       = 0;
		name        = Split-Path -leaf $event.SourceEventArgs.Name;
		nameOld     = "to be set..";
		path        = Split-Path -parent $event.SourceEventArgs.Name;
		pathOld     = "to be set..";
		filePath    = "to be set..";
		filePathOld = "to be set..";
		fullPath    = $event.SourceEventArgs.FullPath;
		fullPathOld = "to be set..";
		included    = (&{if ($tracker.Include.count -eq 0){ $true; } else { $false; }});
		excluded    = $false;
	};

	if ((gm -InputObject $event.SourceEventArgs "OldName") -ne $null){
		$op.nameOld = Split-Path -leaf $event.SourceEventArgs.OldName;
		$op.pathOld = Split-Path -parent $event.SourceEventArgs.OldName;
		$op.fullPathOld = $event.SourceEventArgs.OldFullPath;
	} else {
		$op.nameOld = "";
		$op.pathOld = "";
		$op.fullPathOld = "";
	}

	$op.filePath    = $op.name;
	$op.filePathOld = $op.nameOld;
	if ($op.path){ $op.filePath = $op.path+"\"+$op.name;}
	if ($op.pathOld){ $op.filePathOld = $op.pathOld+"\"+$op.nameOld;}

	for($i=0;$i -lt $tracker.Include.count;$i++){
		if ($tracker.Include[$i].IsMatch($op.filePath) -or
			($op.filePathOld -and $tracker.Include[$i].IsMatch($op.filePathOld) ) ){
			$op.included = $true;
			break;
		}
	}
	if (!$op.included) { return $op; }

	for($i=0;$i -lt $tracker.Exclude.count;$i++){
		if ($tracker.Exclude[$i].IsMatch($op.filePath) -or
			($op.filePathOld -and $tracker.Exclude[$i].IsMatch($op.filePathOld) ) ){
			$op.excluded = $true;
			break;
		}
	}

	return $op;
}
